from dataclasses import dataclass
from enum import Enum
from typing import Literal
from empiric.core.types import RPC_URLS
from starknet_py.net.full_node_client import FullNodeClient

def get_client_from_network(network: str) -> FullNodeClient:
    return FullNodeClient(node_url=RPC_URLS[network])

# contract address configuration
@dataclass
class ContractAddresses:
    publisher_registry_address: int
    oracle_proxy_address: int


TESTNET_CONTRACTS = ContractAddresses(
    0x5CB0AFA98435DE7DA6DA7FAD3B40C9D17E747A57BCA28FEB1C41F05E391F54E,
    0x446812BAC98C08190DEE8967180F4E3CDCD1DB9373CA269904ACB17F67F7093,
)
TESTNET2_CONTRACTS = ContractAddresses(
    0x7DCF59E9146DE058CCB740C8D2C5ACADAAFE3766E12D65C5B95607EDE62317D,
    0xC28F8752ABB9ED18F65FED730B8FAA69BDF6128BB730411EFD916284701938,
)
INTEGRATION_CONTRACTS = ContractAddresses(
    0x062F25D256931E058906542CEE48D645996F0D84C98940AE7D2BD6B68ECC2925,
    0x072D2CD06603E0C957F54A4A1909745939FCFD597EA22323672A6EACBF27D02A,
)
MAINNET_CONTRACTS = ContractAddresses(
    0x04746485FA57B49DC992C35D7F12054B5A7D24B0E187021CD8F40BC2517700BC,
    0x346C57F094D641AD94E43468628D8E9C574DCB2803EC372576CCC60A40BE2C4,
)

from starknet_py.net.models import StarknetChainId
from starkware.python.utils import from_bytes


@dataclass
class Network:
    chain_id: int
    gateway_url: str


class StarknetChainIdExtension(Enum):
    TESTNET2 = from_bytes(b"SN_GOERLI2")
    PRAGMA_TESTNET = from_bytes(b"pragma_goerli")


# network configurations
LOCAL = Network(StarknetChainId.TESTNET, "http://127.0.0.1:5050")
TESTNET = Network(StarknetChainId.TESTNET, "https://alpha4.starknet.io")
TESTNET2 = Network(StarknetChainIdExtension.TESTNET2, "https://alpha4-2.starknet.io")
INTEGRATION = Network(
    StarknetChainId.TESTNET, "https://external.integration.starknet.io"
)
MAINNET = Network(StarknetChainId.MAINNET, "https://alpha-mainnet.starknet.io")
PRAGMA_TESTNET = Network(StarknetChainIdExtension.PRAGMA_TESTNET, "https://testnet.pragmaoracle.com/rpc")

NETWORKS = {
    "testnet": TESTNET,
    "testnet2": TESTNET2,
    "integration": INTEGRATION,
    "mainnet": MAINNET,
    "local": LOCAL,
    "pragma-testnet": PRAGMA_TESTNET,
}

NetworkType = Literal[
    "testnet",
    "testnet2",
    "integration",
    "mainnet",
    "local",
    "pragma-testnet"
]

CONTRACT_ADDRESSES = {
    "testnet": TESTNET_CONTRACTS,
    "testnet2": TESTNET2_CONTRACTS,
    "integration": INTEGRATION_CONTRACTS,
    "mainnet": MAINNET_CONTRACTS,
}

EnvironmentTypes = Literal[
    "testnet",
    "testnet2",
    "integration",
    "mainnet",
    "sharingan",
    "pragma-testnet"
]



import logging
import requests
from aiohttp import ClientSession
from empiric.core.entry import FutureEntry
from empiric.core.utils import currency_pair_to_pair_id
from empiric.publisher.assets import EmpiricAsset, EmpiricFutureAsset
from empiric.publisher.types import PublisherFetchError

logger = logging.getLogger(__name__)

BASE_URL: str = "https://okx.com/api/v5/market/tickers"
TIMESTAMP_URL: str = "https://www.okx.com/api/v5/public/instruments"
SOURCE: str = "OKX"

assets = [
    {"type": "FUTURE", "pair": ("BTC", "USDT"), "decimals": 8},
    {"type": "FUTURE", "pair": ("ETH", "USDT"), "decimals": 8},
]
publisher = 'EMPIRIC'

async def fetch_expiry_timestamp( asset,id,session):
    pair = asset["pair"]
    url = f"{TIMESTAMP_URL}?instType=FUTURES&instId={id}"
    async with session.get(url) as resp:
        if resp.status == 404:
            return PublisherFetchError(
                f"No data found for {'/'.join(pair)} from OKX"
            )
        result = await resp.json(content_type="application/json")
        if (
            result["code"] == "51001"
            or result["msg"] == "Instrument ID does not exist"
        ):
            return PublisherFetchError(
                f"No data found for {'/'.join(pair)} from OKX"
            )
        return result["data"][0]["expTime"]

async def construct_future_entry(asset, result, publisher, session):
    result_arr = []
    result_len = len(result["data"])
    pair = asset["pair"]
    for i in range(0,result_len):
        data = result["data"][i]
        timestamp = int(data["ts"])
        price = float(data["last"])
        price_int = int(price * (10 ** asset["decimals"]))
        pair_id = currency_pair_to_pair_id(*pair)
        volume = float(data["volCcy24h"])
        volume_int = int(volume)
        expiry_timestamp = await fetch_expiry_timestamp(asset, data["instId"], session)
        result_arr.append(FutureEntry(
        pair_id=pair_id,
        price=price_int,
        volume=volume_int,
        timestamp=timestamp,
        source=SOURCE,
        publisher=publisher,
        expiry_timestamp=expiry_timestamp,
    ))
    logger.info(f"Fetched future for {'/'.join(pair)} from OKX")

    return result_arr

async def fetch_future_pair(asset, session, publisher):
    pair = asset["pair"]
    url = f"{BASE_URL}?instType=FUTURES&uly={pair[0]}-{pair[1]}"

    async with session.get(url) as resp:
        if resp.status == 404:
            return PublisherFetchError(
                f"No data found for {'/'.join(pair)} from OKX"
            )
        result = await resp.json(content_type="application/json")
        if (
            result["code"] == "51001"
            or result["msg"] == "Instrument ID does not exist"
        ):
            return PublisherFetchError(
                f"No data found for {'/'.join(pair)} from OKX"
            )
        return await construct_future_entry(asset, result, publisher, session)


async def fetch_futures(assets, session, publisher):
    entries = []
    for asset in assets:
        if asset["type"] != "FUTURE":
            logger.debug(f"Skipping OKX for non-future asset {asset}")
            continue
        future_entries = await fetch_future_pair(asset, session, publisher)
        if isinstance(future_entries, list):
            entries.extend(future_entries)
        else:
            entries.append(future_entries)
    return entries


def print_future_entry(entry):
    print(f"Pair ID: {entry.pair_id}")
    print(f"Price: {entry.price}")
    print(f"Volume: {entry.volume}")
    print(f"Timestamp: {entry.base.timestamp}")
    print(f"Source: {entry.base.source}")
    print(f"Publisher: {entry.base.publisher}")
    print(f"Expiry timestamp: {entry.expiry_timestamp}")
    print("---")

def print_futures(entries):
    for entry in entries:
        if isinstance(entry, PublisherFetchError):
            print(f"Error fetching future: {entry}")
        else:
            print_future_entry(entry)

# Running the async function
import aiohttp
import asyncio

async def run():
    async with aiohttp.ClientSession() as session:
        # Fetch the futures
        entries = await fetch_futures(assets, session, publisher)
        # Print the fetched futures
        print_futures(entries)

asyncio.run(run())



import asyncio
import datetime
import logging
from typing import Dict, List

import requests
from aiohttp import ClientSession
from empiric.core.entry import SpotEntry
from empiric.core.utils import currency_pair_to_pair_id
from empiric.publisher.assets import EmpiricAsset, EmpiricSpotAsset
from empiric.publisher.types import PublisherFetchError, PublisherInterfaceT

logger = logging.getLogger(__name__)

ASSET_MAPPING: Dict[str, str] = {
    "ETH": "ethereum",
    "BTC": "bitcoin",
    "WBTC": "wrapped-bitcoin",
    "SOL": "solana",
    "AVAX": "avalanche-2",
    "DOGE": "dogecoin",
    "SHIB": "shiba-inu",
    "TEMP": "tempus",
    "DAI": "dai",
    "USDT": "tether",
    "USDC": "usd-coin",
    "TUSD": "true-usd",
    "BUSD": "binance-usd",
    "BNB": "binancecoin",
    "ADA": "cardano",
    "XRP": "ripple",
    "MATIC": "matic-network",
    "AAVE": "aave",
    "R": "r",
    "LORDS": "lords",
}


class DefillamaFetcher(PublisherInterfaceT):
    BASE_URL: str = (
        "https://coins.llama.fi/prices/current/coingecko:{pair_id}"
        "?searchWidth=5m"
    )

    SOURCE: str = "DEFILLAMA"
    headers = {
        "Accepts": "application/json",
    }

    publisher: str

    def __init__(self, assets: List[EmpiricAsset], publisher):
        self.assets = assets
        self.publisher = publisher

    async def _fetch_pair(
        self, asset: EmpiricSpotAsset, session: ClientSession
    ) -> SpotEntry:
        pair = asset["pair"]
        pair_id = ASSET_MAPPING.get(pair[0])
        if pair_id is None:
            return PublisherFetchError(
                f"Unknown price pair, do not know how to query Coingecko for {pair[0]}"
            )
        url = self.BASE_URL.format(pair_id=pair_id)

        async with session.get(
            url, headers=self.headers, raise_for_status=True
        ) as resp:
            result = await resp.json()
            return self._construct(asset, result)

    def _fetch_pair_sync(self, asset: EmpiricSpotAsset) -> SpotEntry:
        pair = asset["pair"]
        pair_id = ASSET_MAPPING.get(pair[0])
        if pair_id is None:
            return PublisherFetchError(
                f"Unknown price pair, do not know how to query Coingecko for {pair[0]}"
            )
        url = self.BASE_URL.format(pair_id=pair_id)

        resp = requests.get(url, headers=self.headers)
        resp.raise_for_status()
        result = resp.json()
        return self._construct(asset, result)

    async def fetch(self, session: ClientSession) -> List[SpotEntry]:
        entries = []
        for asset in self.assets:
            if asset["type"] != "SPOT":
                logger.debug(f"Skipping {self.SOURCE} for non-spot asset {asset}")
                continue
            entries.append(asyncio.ensure_future(self._fetch_pair(asset, session)))
        return await asyncio.gather(*entries, return_exceptions=True)

    def fetch_sync(self) -> List[SpotEntry]:
        entries = []
        for asset in self.assets:
            if asset["type"] != "SPOT":
                logger.debug(f"Skipping {self.SOURCE} for non-spot asset {asset}")
                continue
            entries.append(self._fetch_pair_sync(asset))
        return entries

    def _construct(self, asset, result) -> SpotEntry:
        pair = asset["pair"]
        cg_id = ASSET_MAPPING.get(pair[0])
        pair_id = currency_pair_to_pair_id(*pair)
        price = result["coins"][f"coingecko:{cg_id}"]["price"]
        price_int = int(price * (10 ** asset["decimals"]))
        timestamp = int(result["coins"][f"coingecko:{cg_id}"]["timestamp"])

        logger.info(f"Fetched price {price} for {pair_id} from Coingecko")

        return SpotEntry(
            pair_id=pair_id,
            price=price_int,
            timestamp=timestamp,
            source=self.SOURCE,
            publisher=self.publisher,
        )



import asyncio
import logging
import time
from typing import Dict, List

import requests
from aiohttp import ClientSession

from pragma.core.assets import PragmaAsset, PragmaSpotAsset
from pragma.core.entry import SpotEntry
from pragma.core.utils import currency_pair_to_pair_id
from pragma.publisher.types import PublisherFetchError, PublisherInterfaceT

logger = logging.getLogger(__name__)


ASSET_MAPPING: Dict[str, any] = {
    "LORDS": (
        "starknet-alpha",
        "0x124aeb495b947201f5fac96fd1138e326ad86195b98df6dec9009158a533b49",
    ),
    "R": ("eth", "0x183015a9ba6ff60230fdeadc3f43b3d788b13e21"),
    "WBTC": ("eth", "0x2260fac5e5542a773aa44fbcfedf7c193bc2c599"),
}


class GeckoTerminalFetcher(PublisherInterfaceT):
    BASE_URL: str = (
        "https://api.geckoterminal.com/api/v2/networks/{network}/tokens/{token_address}"
    )

    SOURCE: str = "GECKOTERMINAL"
    headers = {
        "Accepts": "application/json",
    }

    publisher: str

    def __init__(self, assets: List[PragmaAsset], publisher):
        self.assets = assets
        self.publisher = publisher

    async def _fetch_pair(
        self, asset: PragmaSpotAsset, session: ClientSession
    ) -> SpotEntry:
        pair = asset["pair"]
        pool = ASSET_MAPPING.get(pair[0])
        if pool is None:
            return PublisherFetchError(
                f"Unknown price pair, do not know how to query GeckoTerminal for {pair[0]}"
            )
        if pair[1] != "USD":
            return PublisherFetchError(f"Base asset not supported : {pair[1]}")
        url = self.BASE_URL.format(network=pool[0], token_address=pool[1])

        async with session.get(url, headers=self.headers) as resp:
            if resp.status == 404:
                return PublisherFetchError(
                    f"No data found for {'/'.join(pair)} from GeckoTerminal"
                )
            result = await resp.json()
            if (
                result.get("errors") is not None
                and result["errors"][0]["title"] == "Not Found"
            ):
                return PublisherFetchError(
                    f"No data found for {'/'.join(pair)} from GeckoTerminal"
                )
            return self._construct(asset, result)

    def _fetch_pair_sync(self, asset: PragmaSpotAsset) -> SpotEntry:
        pair = asset["pair"]
        pool = ASSET_MAPPING.get(pair[0])
        if pool is None:
            return PublisherFetchError(
                f"Unknown price pair, do not know how to query GeckoTerminal for {pair[0]}"
            )
        if pair[1] != "USD":
            return PublisherFetchError(f"Base asset not supported : {pair[1]}")
        url = self.BASE_URL.format(network=pool[0], token_address=pool[1])

        resp = requests.get(url, headers=self.headers)
        if resp.status_code == 404:
            return PublisherFetchError(
                f"No data found for {'/'.join(pair)} from GeckoTerminal"
            )
        result = resp.json()
        if (
            result.get("errors") is not None
            and result["errors"][0]["title"] == "Not Found"
        ):
            return PublisherFetchError(
                f"No data found for {'/'.join(pair)} from GeckoTerminal"
            )
        return self._construct(asset, result)

    async def fetch(self, session: ClientSession) -> List[SpotEntry]:
        entries = []
        for asset in self.assets:
            if asset["type"] != "SPOT":
                logger.debug(f"Skipping {self.SOURCE} for non-spot asset {asset}")
                continue
            entries.append(asyncio.ensure_future(self._fetch_pair(asset, session)))
        return await asyncio.gather(*entries, return_exceptions=True)

    def fetch_sync(self) -> List[SpotEntry]:
        entries = []
        for asset in self.assets:
            if asset["type"] != "SPOT":
                logger.debug(f"Skipping {self.SOURCE} for non-spot asset {asset}")
                continue
            entries.append(self._fetch_pair_sync(asset))
        return entries

    def _construct(self, asset, result) -> SpotEntry:
        pair = asset["pair"]
        data = result["data"]["attributes"]

        price = float(data["price_usd"])
        price_int = int(price * (10 ** asset["decimals"]))
        volume = float(data["volume_usd"]["h24"])

        timestamp = int(time.time())

        pair_id = currency_pair_to_pair_id(*pair)
        logger.info(f"Fetched price {price} for {pair_id} from GeckoTerminal")

        return SpotEntry(
            pair_id=pair_id,
            price=price_int,
            timestamp=timestamp,
            source=self.SOURCE,
            publisher=self.publisher,
            volume=volume,
        )



import abc
from typing import Any, List

import aiohttp
from aiohttp import ClientSession


# Abstract base class for all publishers
class PublisherInterfaceT(abc.ABC):
    @abc.abstractmethod
    async def fetch(self, session: ClientSession) -> List[Any]:
        ...

    @abc.abstractmethod
    def fetch_sync(self) -> List[Any]:
        ...

    async def _fetch(self):
        async with aiohttp.ClientSession() as session:
            data = await self.fetch(session)
            return data


class PublisherFetchError:
    message: str

    def __init__(self, message: str):
        self.message = message

    def serialize(self):
        return self.message

class UnsupportedAssetError:
    message: str

    def __init__(self, message: str):
        self.message = message

    def serialize(self):
        return self.message


ORACLE_ABI = [
    {
        "members": [
            {
                "name": "base",
                "offset": 0,
                "type": "BaseEntry"
            },
            {
                "name": "key",
                "offset": 3,
                "type": "felt"
            },
            {
                "name": "value",
                "offset": 4,
                "type": "felt"
            }
        ],
        "name": "GenericEntry",
        "size": 5,
        "type": "struct"
    },
    {
        "members": [
            {
                "name": "timestamp",
                "offset": 0,
                "type": "felt"
            },
            {
                "name": "source",
                "offset": 1,
                "type": "felt"
            },
            {
                "name": "publisher",
                "offset": 2,
                "type": "felt"
            }
        ],
        "name": "BaseEntry",
        "size": 3,
        "type": "struct"
    },
    {
        "members": [
            {
                "name": "base",
                "offset": 0,
                "type": "BaseEntry"
            },
            {
                "name": "pair_id",
                "offset": 3,
                "type": "felt"
            },
            {
                "name": "price",
                "offset": 4,
                "type": "felt"
            },
            {
                "name": "volume",
                "offset": 5,
                "type": "felt"
            }
        ],
        "name": "SpotEntry",
        "size": 6,
        "type": "struct"
    },
    {
        "members": [
            {
                "name": "base",
                "offset": 0,
                "type": "BaseEntry"
            },
            {
                "name": "pair_id",
                "offset": 3,
                "type": "felt"
            },
            {
                "name": "price",
                "offset": 4,
                "type": "felt"
            },
            {
                "name": "expiry_timestamp",
                "offset": 5,
                "type": "felt"
            },
            {
                "name": "volume",
                "offset": 6,
                "type": "felt"
            }
        ],
        "name": "FutureEntry",
        "size": 7,
        "type": "struct"
    },
    {
        "members": [
            {
                "name": "id",
                "offset": 0,
                "type": "felt"
            },
            {
                "name": "decimals",
                "offset": 1,
                "type": "felt"
            },
            {
                "name": "is_abstract_currency",
                "offset": 2,
                "type": "felt"
            },
            {
                "name": "starknet_address",
                "offset": 3,
                "type": "felt"
            },
            {
                "name": "ethereum_address",
                "offset": 4,
                "type": "felt"
            }
        ],
        "name": "Currency",
        "size": 5,
        "type": "struct"
    },
    {
        "members": [
            {
                "name": "id",
                "offset": 0,
                "type": "felt"
            },
            {
                "name": "quote_currency_id",
                "offset": 1,
                "type": "felt"
            },
            {
                "name": "base_currency_id",
                "offset": 2,
                "type": "felt"
            }
        ],
        "name": "Pair",
        "size": 3,
        "type": "struct"
    },
    {
        "members": [
            {
                "name": "price",
                "offset": 0,
                "type": "felt"
            },
            {
                "name": "decimals",
                "offset": 1,
                "type": "felt"
            },
            {
                "name": "last_updated_timestamp",
                "offset": 2,
                "type": "felt"
            },
            {
                "name": "num_sources_aggregated",
                "offset": 3,
                "type": "felt"
            }
        ],
        "name": "EmpiricPricesResponse",
        "size": 4,
        "type": "struct"
    },
    {
        "members": [
            {
                "name": "timestamp",
                "offset": 0,
                "type": "felt"
            },
            {
                "name": "value",
                "offset": 1,
                "type": "felt"
            },
            {
                "name": "aggregation_mode",
                "offset": 2,
                "type": "felt"
            },
            {
                "name": "num_sources_aggregated",
                "offset": 3,
                "type": "felt"
            }
        ],
        "name": "Checkpoint",
        "size": 4,
        "type": "struct"
    },
    {
        "data": [
            {
                "name": "old_publisher_registry_address",
                "type": "felt"
            },
            {
                "name": "new_publisher_registry_address",
                "type": "felt"
            }
        ],
        "keys": [],
        "name": "UpdatedPublisherRegistryAddress",
        "type": "event"
    },
    {
        "data": [
            {
                "name": "new_entry",
                "type": "GenericEntry"
            }
        ],
        "keys": [],
        "name": "SubmittedEntry",
        "type": "event"
    },
    {
        "data": [
            {
                "name": "new_entry",
                "type": "SpotEntry"
            }
        ],
        "keys": [],
        "name": "SubmittedSpotEntry",
        "type": "event"
    },
    {
        "data": [
            {
                "name": "new_entry",
                "type": "FutureEntry"
            }
        ],
        "keys": [],
        "name": "SubmittedFutureEntry",
        "type": "event"
    },
    {
        "data": [
            {
                "name": "currency",
                "type": "Currency"
            }
        ],
        "keys": [],
        "name": "SubmittedCurrency",
        "type": "event"
    },
    {
        "data": [
            {
                "name": "currency",
                "type": "Currency"
            }
        ],
        "keys": [],
        "name": "UpdatedCurrency",
        "type": "event"
    },
    {
        "data": [
            {
                "name": "pair",
                "type": "Pair"
            }
        ],
        "keys": [],
        "name": "SubmittedPair",
        "type": "event"
    },
    {
        "data": [
            {
                "name": "pair_id",
                "type": "felt"
            }
        ],
        "keys": [],
        "name": "CheckpointSpotEntry",
        "type": "event"
    },
    {
        "data": [
            {
                "name": "pair_id",
                "type": "felt"
            },
            {
                "name": "expiry_timestamp",
                "type": "felt"
            }
        ],
        "keys": [],
        "name": "CheckpointFutureEntry",
        "type": "event"
    },
    {
        "data": [
            {
                "name": "implementation",
                "type": "felt"
            }
        ],
        "keys": [],
        "name": "Upgraded",
        "type": "event"
    },
    {
        "data": [
            {
                "name": "previousAdminAddress",
                "type": "felt"
            },
            {
                "name": "newAdminAddress",
                "type": "felt"
            }
        ],
        "keys": [],
        "name": "AdminAddressChanged",
        "type": "event"
    },
    {
        "inputs": [
            {
                "name": "proxy_admin",
                "type": "felt"
            },
            {
                "name": "publisher_registry_address",
                "type": "felt"
            },
            {
                "name": "currencies_len",
                "type": "felt"
            },
            {
                "name": "currencies",
                "type": "Currency*"
            },
            {
                "name": "pairs_len",
                "type": "felt"
            },
            {
                "name": "pairs",
                "type": "Pair*"
            }
        ],
        "name": "initializer",
        "outputs": [],
        "type": "function"
    },
    {
        "inputs": [
            {
                "name": "pair_id",
                "type": "felt"
            },
            {
                "name": "sources_len",
                "type": "felt"
            },
            {
                "name": "sources",
                "type": "felt*"
            }
        ],
        "name": "get_spot_entries_for_sources",
        "outputs": [
            {
                "name": "entries_len",
                "type": "felt"
            },
            {
                "name": "entries",
                "type": "SpotEntry*"
            }
        ],
        "stateMutability": "view",
        "type": "function"
    },
    {
        "inputs": [
            {
                "name": "pair_id",
                "type": "felt"
            }
        ],
        "name": "get_spot_entries",
        "outputs": [
            {
                "name": "entries_len",
                "type": "felt"
            },
            {
                "name": "entries",
                "type": "SpotEntry*"
            }
        ],
        "stateMutability": "view",
        "type": "function"
    },
    {
        "inputs": [
            {
                "name": "pair_id",
                "type": "felt"
            },
            {
                "name": "source",
                "type": "felt"
            }
        ],
        "name": "get_spot_entry",
        "outputs": [
            {
                "name": "entry",
                "type": "SpotEntry"
            }
        ],
        "stateMutability": "view",
        "type": "function"
    },
    {
        "inputs": [
            {
                "name": "pair_id",
                "type": "felt"
            },
            {
                "name": "expiry_timestamp",
                "type": "felt"
            },
            {
                "name": "sources_len",
                "type": "felt"
            },
            {
                "name": "sources",
                "type": "felt*"
            }
        ],
        "name": "get_future_entries_for_sources",
        "outputs": [
            {
                "name": "entries_len",
                "type": "felt"
            },
            {
                "name": "entries",
                "type": "FutureEntry*"
            }
        ],
        "stateMutability": "view",
        "type": "function"
    },
    {
        "inputs": [
            {
                "name": "pair_id",
                "type": "felt"
            },
            {
                "name": "expiry_timestamp",
                "type": "felt"
            }
        ],
        "name": "get_future_entries",
        "outputs": [
            {
                "name": "entries_len",
                "type": "felt"
            },
            {
                "name": "entries",
                "type": "FutureEntry*"
            }
        ],
        "stateMutability": "view",
        "type": "function"
    },
    {
        "inputs": [
            {
                "name": "pair_id",
                "type": "felt"
            },
            {
                "name": "expiry_timestamp",
                "type": "felt"
            },
            {
                "name": "source",
                "type": "felt"
            }
        ],
        "name": "get_future_entry",
        "outputs": [
            {
                "name": "entry",
                "type": "FutureEntry"
            }
        ],
        "stateMutability": "view",
        "type": "function"
    },
    {
        "inputs": [
            {
                "name": "pair_id",
                "type": "felt"
            }
        ],
        "name": "get_spot_median",
        "outputs": [
            {
                "name": "price",
                "type": "felt"
            },
            {
                "name": "decimals",
                "type": "felt"
            },
            {
                "name": "last_updated_timestamp",
                "type": "felt"
            },
            {
                "name": "num_sources_aggregated",
                "type": "felt"
            }
        ],
        "stateMutability": "view",
        "type": "function"
    },
    {
        "inputs": [
            {
                "name": "pair_id",
                "type": "felt"
            },
            {
                "name": "sources_len",
                "type": "felt"
            },
            {
                "name": "sources",
                "type": "felt*"
            }
        ],
        "name": "get_spot_median_for_sources",
        "outputs": [
            {
                "name": "price",
                "type": "felt"
            },
            {
                "name": "decimals",
                "type": "felt"
            },
            {
                "name": "last_updated_timestamp",
                "type": "felt"
            },
            {
                "name": "num_sources_aggregated",
                "type": "felt"
            }
        ],
        "stateMutability": "view",
        "type": "function"
    },
    {
        "inputs": [
            {
                "name": "pair_ids_len",
                "type": "felt"
            },
            {
                "name": "pair_ids",
                "type": "felt*"
            },
            {
                "name": "idx",
                "type": "felt"
            }
        ],
        "name": "get_spot_median_multi",
        "outputs": [
            {
                "name": "prices_response_len",
                "type": "felt"
            },
            {
                "name": "prices_response",
                "type": "EmpiricPricesResponse*"
            }
        ],
        "stateMutability": "view",
        "type": "function"
    },
    {
        "inputs": [
            {
                "name": "pair_ids_len",
                "type": "felt"
            },
            {
                "name": "pair_ids",
                "type": "felt*"
            },
            {
                "name": "idx",
                "type": "felt"
            },
            {
                "name": "prices_response_len",
                "type": "felt"
            },
            {
                "name": "prices_response",
                "type": "EmpiricPricesResponse*"
            }
        ],
        "name": "get_spot_median_multi_loop",
        "outputs": [],
        "stateMutability": "view",
        "type": "function"
    },
    {
        "inputs": [
            {
                "name": "pair_id",
                "type": "felt"
            },
            {
                "name": "aggregation_mode",
                "type": "felt"
            }
        ],
        "name": "get_spot",
        "outputs": [
            {
                "name": "price",
                "type": "felt"
            },
            {
                "name": "decimals",
                "type": "felt"
            },
            {
                "name": "last_updated_timestamp",
                "type": "felt"
            },
            {
                "name": "num_sources_aggregated",
                "type": "felt"
            }
        ],
        "stateMutability": "view",
        "type": "function"
    },
    {
        "inputs": [
            {
                "name": "pair_id",
                "type": "felt"
            },
            {
                "name": "aggregation_mode",
                "type": "felt"
            },
            {
                "name": "sources_len",
                "type": "felt"
            },
            {
                "name": "sources",
                "type": "felt*"
            }
        ],
        "name": "get_spot_for_sources",
        "outputs": [
            {
                "name": "price",
                "type": "felt"
            },
            {
                "name": "decimals",
                "type": "felt"
            },
            {
                "name": "last_updated_timestamp",
                "type": "felt"
            },
            {
                "name": "num_sources_aggregated",
                "type": "felt"
            }
        ],
        "stateMutability": "view",
        "type": "function"
    },
    {
        "inputs": [],
        "name": "get_publisher_registry_address",
        "outputs": [
            {
                "name": "publisher_registry_address",
                "type": "felt"
            }
        ],
        "stateMutability": "view",
        "type": "function"
    },
    {
        "inputs": [
            {
                "name": "pair_id",
                "type": "felt"
            }
        ],
        "name": "get_spot_decimals",
        "outputs": [
            {
                "name": "decimals",
                "type": "felt"
            }
        ],
        "stateMutability": "view",
        "type": "function"
    },
    {
        "inputs": [
            {
                "name": "key",
                "type": "felt"
            }
        ],
        "name": "get_value",
        "outputs": [
            {
                "name": "value",
                "type": "felt"
            },
            {
                "name": "decimals",
                "type": "felt"
            },
            {
                "name": "last_updated_timestamp",
                "type": "felt"
            },
            {
                "name": "num_sources_aggregated",
                "type": "felt"
            }
        ],
        "stateMutability": "view",
        "type": "function"
    },
    {
        "inputs": [
            {
                "name": "pair_id",
                "type": "felt"
            },
            {
                "name": "expiry_timestamp",
                "type": "felt"
            },
            {
                "name": "aggregation_mode",
                "type": "felt"
            }
        ],
        "name": "get_futures",
        "outputs": [
            {
                "name": "value",
                "type": "felt"
            },
            {
                "name": "decimals",
                "type": "felt"
            },
            {
                "name": "last_updated_timestamp",
                "type": "felt"
            },
            {
                "name": "num_sources_aggregated",
                "type": "felt"
            }
        ],
        "stateMutability": "view",
        "type": "function"
    },
    {
        "inputs": [
            {
                "name": "base_currency_id",
                "type": "felt"
            },
            {
                "name": "quote_currency_id",
                "type": "felt"
            },
            {
                "name": "aggregation_mode",
                "type": "felt"
            }
        ],
        "name": "get_spot_with_USD_hop",
        "outputs": [
            {
                "name": "price",
                "type": "felt"
            },
            {
                "name": "decimals",
                "type": "felt"
            },
            {
                "name": "last_updated_timestamp",
                "type": "felt"
            },
            {
                "name": "num_sources_aggregated",
                "type": "felt"
            }
        ],
        "stateMutability": "view",
        "type": "function"
    },
    {
        "inputs": [
            {
                "name": "currency_ids_len",
                "type": "felt"
            },
            {
                "name": "currency_ids",
                "type": "felt*"
            },
            {
                "name": "aggregation_mode",
                "type": "felt"
            },
            {
                "name": "idx",
                "type": "felt"
            },
            {
                "name": "price",
                "type": "felt"
            },
            {
                "name": "decimals",
                "type": "felt"
            },
            {
                "name": "last_updated_timestamp",
                "type": "felt"
            },
            {
                "name": "num_sources_aggregated",
                "type": "felt"
            }
        ],
        "name": "get_spot_with_hop",
        "outputs": [
            {
                "name": "price",
                "type": "felt"
            },
            {
                "name": "decimals",
                "type": "felt"
            },
            {
                "name": "last_updated_timestamp",
                "type": "felt"
            },
            {
                "name": "num_sources_aggregated",
                "type": "felt"
            }
        ],
        "stateMutability": "view",
        "type": "function"
    },
    {
        "inputs": [
            {
                "name": "key",
                "type": "felt"
            },
            {
                "name": "source",
                "type": "felt"
            }
        ],
        "name": "get_entry",
        "outputs": [
            {
                "name": "entry",
                "type": "GenericEntry"
            }
        ],
        "stateMutability": "view",
        "type": "function"
    },
    {
        "inputs": [
            {
                "name": "key",
                "type": "felt"
            }
        ],
        "name": "get_entries",
        "outputs": [
            {
                "name": "entries_len",
                "type": "felt"
            },
            {
                "name": "entries",
                "type": "GenericEntry*"
            }
        ],
        "stateMutability": "view",
        "type": "function"
    },
    {
        "inputs": [
            {
                "name": "key",
                "type": "felt"
            },
            {
                "name": "sources_len",
                "type": "felt"
            },
            {
                "name": "sources",
                "type": "felt*"
            }
        ],
        "name": "get_entries_for_sources",
        "outputs": [
            {
                "name": "entries_len",
                "type": "felt"
            },
            {
                "name": "entries",
                "type": "GenericEntry*"
            }
        ],
        "stateMutability": "view",
        "type": "function"
    },
    {
        "inputs": [
            {
                "name": "new_entry",
                "type": "FutureEntry"
            }
        ],
        "name": "publish_future_entry",
        "outputs": [],
        "type": "function"
    },
    {
        "inputs": [
            {
                "name": "new_entry",
                "type": "SpotEntry"
            }
        ],
        "name": "publish_spot_entry",
        "outputs": [],
        "type": "function"
    },
    {
        "inputs": [
            {
                "name": "new_entry",
                "type": "GenericEntry"
            }
        ],
        "name": "publish_entry",
        "outputs": [],
        "type": "function"
    },
    {
        "inputs": [
            {
                "name": "new_entries_len",
                "type": "felt"
            },
            {
                "name": "new_entries",
                "type": "GenericEntry*"
            }
        ],
        "name": "publish_entries",
        "outputs": [],
        "type": "function"
    },
    {
        "inputs": [
            {
                "name": "new_entries_len",
                "type": "felt"
            },
            {
                "name": "new_entries",
                "type": "FutureEntry*"
            }
        ],
        "name": "publish_future_entries",
        "outputs": [],
        "type": "function"
    },
    {
        "inputs": [
            {
                "name": "new_entries_len",
                "type": "felt"
            },
            {
                "name": "new_entries",
                "type": "SpotEntry*"
            }
        ],
        "name": "publish_spot_entries",
        "outputs": [],
        "type": "function"
    },
    {
        "inputs": [
            {
                "name": "publisher_registry_address",
                "type": "felt"
            }
        ],
        "name": "update_publisher_registry_address",
        "outputs": [],
        "type": "function"
    },
    {
        "inputs": [
            {
                "name": "currency",
                "type": "Currency"
            }
        ],
        "name": "add_currency",
        "outputs": [],
        "type": "function"
    },
    {
        "inputs": [
            {
                "name": "currency",
                "type": "Currency"
            }
        ],
        "name": "update_currency",
        "outputs": [],
        "type": "function"
    },
    {
        "inputs": [
            {
                "name": "pair",
                "type": "Pair"
            }
        ],
        "name": "add_pair",
        "outputs": [],
        "type": "function"
    },
    {
        "inputs": [
            {
                "name": "key",
                "type": "felt"
            }
        ],
        "name": "get_latest_checkpoint_index",
        "outputs": [
            {
                "name": "latest",
                "type": "felt"
            }
        ],
        "stateMutability": "view",
        "type": "function"
    },
    {
        "inputs": [
            {
                "name": "key",
                "type": "felt"
            },
            {
                "name": "expiry_timestamp",
                "type": "felt"
            }
        ],
        "name": "get_latest_future_checkpoint_index",
        "outputs": [
            {
                "name": "latest",
                "type": "felt"
            }
        ],
        "stateMutability": "view",
        "type": "function"
    },
    {
        "inputs": [
            {
                "name": "key",
                "type": "felt"
            },
            {
                "name": "index",
                "type": "felt"
            }
        ],
        "name": "get_checkpoint",
        "outputs": [
            {
                "name": "checkpoint",
                "type": "Checkpoint"
            }
        ],
        "stateMutability": "view",
        "type": "function"
    },
    {
        "inputs": [
            {
                "name": "key",
                "type": "felt"
            },
            {
                "name": "expiry_timestamp",
                "type": "felt"
            },
            {
                "name": "index",
                "type": "felt"
            }
        ],
        "name": "get_future_checkpoint",
        "outputs": [
            {
                "name": "checkpoint",
                "type": "Checkpoint"
            }
        ],
        "stateMutability": "view",
        "type": "function"
    },
    {
        "inputs": [],
        "name": "get_sources_threshold",
        "outputs": [
            {
                "name": "threshold",
                "type": "felt"
            }
        ],
        "stateMutability": "view",
        "type": "function"
    },
    {
        "inputs": [
            {
                "name": "new_implementation",
                "type": "felt"
            }
        ],
        "name": "upgrade",
        "outputs": [],
        "type": "function"
    },
    {
        "inputs": [
            {
                "name": "new_admin_address",
                "type": "felt"
            }
        ],
        "name": "set_admin_address",
        "outputs": [],
        "type": "function"
    },
    {
        "inputs": [],
        "name": "get_implementation_hash",
        "outputs": [
            {
                "name": "address",
                "type": "felt"
            }
        ],
        "stateMutability": "view",
        "type": "function"
    },
    {
        "inputs": [],
        "name": "get_admin_address",
        "outputs": [
            {
                "name": "admin_address",
                "type": "felt"
            }
        ],
        "stateMutability": "view",
        "type": "function"
    },
    {
        "inputs": [
            {
                "name": "pair_id",
                "type": "felt"
            },
            {
                "name": "aggregation_mode",
                "type": "felt"
            }
        ],
        "name": "set_checkpoint",
        "outputs": [],
        "type": "function"
    },
    {
        "inputs": [
            {
                "name": "pair_ids_len",
                "type": "felt"
            },
            {
                "name": "pair_ids",
                "type": "felt*"
            },
            {
                "name": "aggregation_mode",
                "type": "felt"
            }
        ],
        "name": "set_checkpoints",
        "outputs": [],
        "type": "function"
    },
    {
        "inputs": [
            {
                "name": "pair_id",
                "type": "felt"
            },
            {
                "name": "expiry_timestamp",
                "type": "felt"
            },
            {
                "name": "aggregation_mode",
                "type": "felt"
            }
        ],
        "name": "set_future_checkpoint",
        "outputs": [],
        "type": "function"
    },
    {
        "inputs": [
            {
                "name": "pair_id",
                "type": "felt"
            },
            {
                "name": "timestamp",
                "type": "felt"
            }
        ],
        "name": "get_last_spot_checkpoint_before",
        "outputs": [
            {
                "name": "checkpoint",
                "type": "Checkpoint"
            },
            {
                "name": "idx",
                "type": "felt"
            }
        ],
        "stateMutability": "view",
        "type": "function"
    },
    {
        "inputs": [
            {
                "name": "pair_id",
                "type": "felt"
            },
            {
                "name": "expiry_timestamp",
                "type": "felt"
            },
            {
                "name": "timestamp",
                "type": "felt"
            }
        ],
        "name": "get_last_future_checkpoint_before",
        "outputs": [
            {
                "name": "checkpoint",
                "type": "Checkpoint"
            },
            {
                "name": "idx",
                "type": "felt"
            }
        ],
        "stateMutability": "view",
        "type": "function"
    },
    {
        "inputs": [
            {
                "name": "threshold",
                "type": "felt"
            }
        ],
        "name": "set_sources_threshold",
        "outputs": [],
        "type": "function"
    }
]


import logging
from typing import List, Union
import asyncio
import requests
from aiohttp import ClientSession
from empiric.core.entry import FutureEntry
from empiric.core.utils import currency_pair_to_pair_id
from empiric.publisher.assets import EmpiricAsset, EmpiricFutureAsset
from empiric.publisher.types import PublisherFetchError, PublisherInterfaceT

logger = logging.getLogger(__name__)


class OkxFutureFetcher(PublisherInterfaceT):
    BASE_URL: str = "https://okx.com/api/v5/market/tickers"
    SOURCE: str = "OKX"
    TIMESTAMP_URL: str = "https://www.okx.com/api/v5/public/instruments"
    publisher: str

    def __init__(self, assets: List[EmpiricAsset], publisher):
        self.assets = assets
        self.publisher = publisher

    async def fetch_expiry_timestamp(self, asset, id, session):
        pair = asset["pair"]
        url = f"{self.TIMESTAMP_URL}?instType=FUTURES&instId={id}"
        async with session.get(url) as resp:
            if resp.status == 404:
                return PublisherFetchError(
                    f"No data found for {'/'.join(pair)} from OKX"
                )
            result = await resp.json(content_type="application/json")
            if (
                result["code"] == "51001"
                or result["msg"] == "Instrument ID does not exist"
            ):
                return PublisherFetchError(
                    f"No data found for {'/'.join(pair)} from OKX"
                )
            return result["data"][0]["expTime"]

    def fetch_sync_expiry_timestamp(self, asset, id):
        pair = asset["pair"]
        url = f"{self.TIMESTAMP_URL}?instType=FUTURES&instId={id}"
        resp = requests.get(url)
        if resp.status_code == 404:
            return PublisherFetchError(
                f"No data found for {'/'.join(pair)} from OKX"
            )
        result = resp.json()
        if (
            result["code"] == "51001"
            or result["msg"] == "Instrument ID does not exist"
        ):
            return PublisherFetchError(
                f"No data found for {'/'.join(pair)} from OKX"
            )
        return result["data"][0]["expTime"]
    def _construct(self, asset, data, expiry_timestamp) -> List[FutureEntry]:
    
        pair = asset["pair"]
        timestamp = int(int(data["ts"]) / 1000)
        price = float(data["last"])
        price_int = int(price * (10 ** asset["decimals"]))
        pair_id = currency_pair_to_pair_id(*pair)
        volume = float(data["volCcy24h"])
        volume_int = int(volume)
        logger.info(f"Fetched future for {'/'.join(pair)} from OKX")

        return FutureEntry(
                pair_id=pair_id,
                price=price_int,
                volume=volume_int,
                timestamp=timestamp,
                source=self.SOURCE,
                publisher=self.publisher,
                expiry_timestamp=int(expiry_timestamp),
            )

    async def _fetch_pair(self, asset: EmpiricFutureAsset, session: ClientSession):
        pair = asset["pair"]
        url = f"{self.BASE_URL}?instType=FUTURES&uly={pair[0]}-{pair[1]}"
        future_entries = []
        async with session.get(url) as resp:
            if resp.status == 404:
                return PublisherFetchError(
                    f"No data found for {'/'.join(pair)} from OKX"
                )
            result = await resp.json(content_type="application/json")
            if (
                result["code"] == "51001"
                or result["msg"] == "Instrument ID does not exist"
            ):
                return PublisherFetchError(
                    f"No data found for {'/'.join(pair)} from OKX"
                )
            result_len = len(result["data"])
            if result_len > 1:
                for i in range(0, result_len):
                    expiry_timestamp = await self.fetch_expiry_timestamp(asset, result["data"][i]["instId"], session)
                    future_entries.append(self._construct(asset, result['data'][i],expiry_timestamp))
            return future_entries
    
    def _fetch_pair_sync(
        self, asset: EmpiricFutureAsset
    ) -> Union[FutureEntry, PublisherFetchError]:
        pair = asset["pair"]
        future_entries = []
        url = f"{self.BASE_URL}?instType=FUTURES&uly={pair[0]}-{pair[1]}"

        resp = requests.get(url)
        if resp.status_code == 404:
            return PublisherFetchError(f"No data found for {'/'.join(pair)} from OKX")
        result = resp.json(content_type="text/json")
        if result["code"] == "51001" or result["msg"] == "Instrument ID does not exist":
            return PublisherFetchError(f"No data found for {'/'.join(pair)} from OKX")
        result_len = len(result["data"])
        if result_len > 1:
            for i in range(0, result_len):
                expiry_timestamp = self.fetch_sync_expiry_timestamp(asset, result["data"][i]["instId"])
                future_entries.append(self._construct(asset, result['data'][i],expiry_timestamp))
        return future_entries

    def fetch_sync(self):
        entries = []
        for asset in self.assets:
            if asset["type"] != "FUTURE":
                logger.debug(f"Skipping OKX for non-future asset {asset}")
                continue
            future_entries = self._fetch_pair_sync(asset)
            if isinstance(future_entries, list):
                entries.extend(future_entries)
            else:
                entries.append(future_entries)
        return entries

    async def fetch(self, session: ClientSession):
        entries = []
        for asset in self.assets:
            if asset["type"] != "FUTURE":
                logger.debug(f"Skipping OKX for non-future asset {asset}")
                continue
            future_entries = await self._fetch_pair(asset, session)
            if isinstance(future_entries, list):
                entries.extend(future_entries)
            else:
                entries.append(future_entries)
        return entries

    def print_future_entry(entry):
        print(f"Pair ID: {entry.pair_id}")
        print(f"Price: {entry.price}")
        print(f"Volume: {entry.volume}")
        print(f"Timestamp: {entry.base.timestamp}")
        print(f"Source: {entry.base.source}")
        print(f"Publisher: {entry.base.publisher}")
        print(f"Expiry timestamp: {entry.expiry_timestamp}")
        print("---")

    def print_futures(self, entries):
        for entry in entries:
            if isinstance(entry, PublisherFetchError):
                print(f"Error fetching future: {entry}")
            else:
                self.print_future_entry(entry)



import logging

logger = logging.getLogger(__name__)


def get_stream_logger():
    import sys

    logger.setLevel(logging.DEBUG)
    streamHandler = logging.StreamHandler(sys.stdout)
    formatter = logging.Formatter(
        "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    )
    streamHandler.setFormatter(formatter)
    logger.addHandler(streamHandler)

    return logger



from typing import Dict, Literal, Optional, Union

from starknet_py.net.client import Client
from starknet_py.net.client_models import TransactionReceipt, TransactionStatus


class NonceMixin:
    client: Client
    account_contract_address: Optional[int]
    nonce_status: Dict[int, TransactionStatus] = dict()
    nonce_dict: Dict[int, int] = dict()
    pending_nonce: Optional[int] = None

    async def _get_nonce(self) -> int:
        self.update_nonce_dict()
        self.cleanup_nonce_dict()

        if self.pending_nonce:
            return self.pending_nonce
        if self.nonce_status:
            return max(self.nonce_status) + 1

        latest_nonce = await self.get_nonce()
        return latest_nonce

    def cleanup_nonce_dict(self):
        nonce_seq = [
            x
            for x in self.nonce_dict
            if self.nonce_status[x]
            in [
                TransactionStatus.ACCEPTED_ON_L1,
                TransactionStatus.ACCEPTED_ON_L2,
                TransactionStatus.REJECTED,
            ]
        ]
        if nonce_seq:
            max_accepted = max(nonce_seq)
            for nonce in list(self.nonce_dict):
                if nonce <= max_accepted:
                    del self.nonce_dict[nonce]
                    if nonce in self.nonce_status:
                        del self.nonce_status[nonce]
            if self.pending_nonce and self.pending_nonce < max_accepted:
                self.pending_nonce = None

    async def track_nonce(
        self,
        nonce,
        transaction_hash,
    ):
        self.nonce_dict[nonce] = transaction_hash
        self.nonce_status[nonce] = await self.get_status(transaction_hash)

        nonce_min = min(self.nonce_dict)
        nonce_max = max(self.nonce_dict)
        for i in range(nonce_min, nonce_max + 1):
            if i not in self.nonce_dict:
                self.pending_nonce = i
                break

    async def update_nonce_dict(
        self,
    ):
        for nonce in list(self.nonce_dict):
            self.nonce_status[nonce] = await self.get_status(self.nonce_dict[nonce])
            if self.nonce_dict[nonce] in [
                TransactionStatus.REJECTED,
            ]:
                # assume all later transaction will fail because this nonce was skipped
                self.pending_nonce = nonce
                self.nonce_dict = {}
                self.nonce_status = {}
                return

    async def get_nonce(
        self,
        include_pending=True,
        block_number: Optional[Union[int, str, Literal["pending", "latest"]]] = None,
    ):
        if not block_number:
            block_number = "pending" if include_pending else "latest"

        nonce = await self.client.get_contract_nonce(
            self.account_contract_address,
            block_number=block_number,
        )
        return nonce

    async def get_status(
        self,
        transaction_hash: int,
    ) -> TransactionReceipt:
        receipt = await self.client.get_transaction_receipt(transaction_hash)
        return receipt.status



import asyncio
import logging
from typing import List, Union

import requests
from aiohttp import ClientSession
from empiric.core.entry import FutureEntry
from empiric.core.utils import currency_pair_to_pair_id
from empiric.publisher.assets import EmpiricAsset, EmpiricFutureAsset
from empiric.publisher.types import PublisherFetchError, PublisherInterfaceT

logger = logging.getLogger(__name__)


class ByBitFutureFetcher(PublisherInterfaceT):
    BASE_URL: str = "https://api.bybit.com/derivatives/v3/public/tickers?category=linear&symbol="
    SOURCE: str = "BYBIT"

    publisher: str

    def __init__(self, assets: List[EmpiricAsset], publisher):
        self.assets = assets
        self.publisher = publisher

    async def _fetch_pair(
        self, asset: EmpiricFutureAsset, session: ClientSession
    ) -> Union[FutureEntry, PublisherFetchError]:
        pair = asset["pair"]
        url = f"{self.BASE_URL}{pair[0]}{pair[1]}"

        async with session.get(url) as resp:
            if resp.status == 404:
                return PublisherFetchError(
                    f"No data found for {'/'.join(pair)} from BYBIT"
                )
            result = await resp.json(content_type="application/json")
            if (
                result["retCode"] == "51001"
                or result["retMsg"] == "Instrument ID does not exist"
            ):
                return PublisherFetchError(
                    f"No data found for {'/'.join(pair)} from BYBIT"
                )

            return self._construct(asset, result)

    def _fetch_pair_sync(
        self, asset: EmpiricFutureAsset
    ) -> Union[FutureEntry, PublisherFetchError]:
        pair = asset["pair"]
        url = f"{self.BASE_URL}{pair[0]}{pair[1]}"

        resp = requests.get(url)
        if resp.status_code == 404:
            return PublisherFetchError(f"No data found for {'/'.join(pair)} from BYBIT")
        result = resp.json(content_type="application/json")
        if result["retCode"] == "51001" or result["retMsg"] == "Instrument ID does not exist":
            return PublisherFetchError(f"No data found for {'/'.join(pair)} from BYBIT")

        return self._construct(asset, result)

    async def fetch(
        self, session: ClientSession
    ) -> List[Union[FutureEntry, PublisherFetchError]]:
        entries = []
        for asset in self.assets:
            if asset["type"] != "FUTURE":
                logger.debug(f"Skipping BYBIT for non-spot asset {asset}")
                continue
            entries.append(asyncio.ensure_future(self._fetch_pair(asset, session)))
        return await asyncio.gather(*entries, return_exceptions=True)

    def fetch_sync(self) -> List[Union[FutureEntry, PublisherFetchError]]:
        entries = []
        for asset in self.assets:
            if asset["type"] != "FUTURE":
                logger.debug(f"Skipping BYBIT for non-spot asset {asset}")
                continue
            entries.append(self._fetch_pair_sync(asset))
        return entries

    def _construct(self, asset, result) -> FutureEntry:
        pair = asset["pair"]
        data = result["result"]["list"][0]
        timestamp = int(int(result["time"]) / 1000)
        price = float(data["lastPrice"])
        price_int = int(price * (10 ** asset["decimals"]))
        pair_id = currency_pair_to_pair_id(*pair)
        volume = float(data["volume24h"])
        volume_int = int(volume)
        expiry_timestamp = int(data["deliveryTime"])
        logger.info(f"Fetched future for {'/'.join(pair)} from BYBIT")

        return FutureEntry(
        pair_id=pair_id,
        price=price_int,
        volume=volume_int,
        timestamp=timestamp,
        source=self.SOURCE,
        publisher=self.publisher,
        expiry_timestamp=expiry_timestamp,
    )


import logging
from typing import Optional

from empiric.core.abis import ORACLE_ABI, PUBLISHER_REGISTRY_ABI, SUMMARY_STATS_ABI
from empiric.core.config import CONTRACT_ADDRESSES, NETWORKS, ContractAddresses, get_client_from_network
from empiric.core.contract import Contract
from empiric.core.mixins import (
    NonceMixin,
    OracleMixin,
    PublisherRegistryMixin,
    RandomnessMixin,
    TransactionMixin,
)
from starknet_py.net.account.account import Account
from starknet_py.net.signer.stark_curve_signer import KeyPair, StarkCurveSigner

logger = logging.getLogger(__name__)

class EmpiricClient(
    NonceMixin, OracleMixin, PublisherRegistryMixin, RandomnessMixin, TransactionMixin
):
    is_user_client: bool = False
    account_contract_address: Optional[int] = None
    account: Account = None

    def __init__(
        self,
        network: str = "testnet",
        account_private_key: Optional[int] = None,
        account_contract_address: Optional[int] = None,
        contract_addresses_config: Optional[ContractAddresses] = None,
    ):
        """
        Client for interacting with Empiric on Starknet.
        :param net: Target network for the client. Can be a string with URL, one of ``"mainnet"``, ``"testnet"``, ``"local"`` or ``"integration""``
        :param account_private_key: Optional private key for requests.  Not necessary if not making network updates
        :param account_contract_address: Optional account contract address.  Not necessary if not making network updates
        :param contract_addresses_config: Optional Contract Addresses for Empiric.  Will default to the provided network but must be set if using non standard contracts.
        """
        self.network_config = NETWORKS[network]
        self.network = network

        self.client = get_client_from_network(network)

        if account_contract_address and account_private_key:
            self._setup_account_client(
                self.network_config.chain_id,
                account_private_key,
                account_contract_address,
            )

        if not contract_addresses_config:
            contract_addresses_config = CONTRACT_ADDRESSES[network]
        self.contract_addresses_config = contract_addresses_config
        self._setup_contracts()

    def _setup_contracts(self):
        provider = self.account if self.account else self.client
        self.oracle = Contract(
            address=self.contract_addresses_config.oracle_proxy_address,
            abi=ORACLE_ABI,
            provider=provider,
        )
        self.publisher_registry = Contract(
            address=self.contract_addresses_config.publisher_registry_address,
            abi=PUBLISHER_REGISTRY_ABI,
            provider=provider,
        )

    async def get_balance(self, account_contract_address, token_address=None):
        client = Account(
            address=account_contract_address,
            client=self.client,
            key_pair=KeyPair.from_private_key(1),
            chain=self.network_config.chain_id,
        )
        balance = await client.get_balance(token_address)
        return balance

    def set_account(self, chain_id, private_key, account_contract_address):
        self._setup_account_client(chain_id, private_key, account_contract_address)

    def _setup_account_client(self, chain_id, private_key, account_contract_address):
        self.signer = StarkCurveSigner(
            account_contract_address,
            KeyPair.from_private_key(private_key),
            chain_id,
        )
        self.account = Account(
            address=account_contract_address,
            client=self.client,
            signer=self.signer,
        )
        self.client = self.account.client
        self.client._get_nonce = self._get_nonce
        self.is_user_client = True
        self.account_contract_address = account_contract_address

    def account_address(self):
        return self.account.address

    def init_stats_contract(
        self,
        stats_contract_address: int,
    ):
        provider = self.account if self.account else self.client
        self.stats = Contract(
            address=stats_contract_address,
            abi=SUMMARY_STATS_ABI,
            provider=provider,
        )



RANDOMNESS_ABI = [
    {
        "members": [
            {"name": "low", "offset": 0, "type": "felt"},
            {"name": "high", "offset": 1, "type": "felt"},
        ],
        "name": "Uint256",
        "size": 2,
        "type": "struct",
    },
    {
        "data": [{"name": "implementation", "type": "felt"}],
        "keys": [],
        "name": "Upgraded",
        "type": "event",
    },
    {
        "data": [
            {"name": "previousAdminAddress", "type": "felt"},
            {"name": "newAdminAddress", "type": "felt"},
        ],
        "keys": [],
        "name": "AdminAddressChanged",
        "type": "event",
    },
    {
        "data": [
            {"name": "request_id", "type": "felt"},
            {"name": "caller_address", "type": "felt"},
            {"name": "seed", "type": "felt"},
            {"name": "minimum_block_number", "type": "felt"},
            {"name": "callback_address", "type": "felt"},
            {"name": "callback_gas_limit", "type": "felt"},
            {"name": "num_words", "type": "felt"},
        ],
        "keys": [],
        "name": "Randomness__request",
        "type": "event",
    },
    {
        "data": [
            {"name": "request_id", "type": "felt"},
            {"name": "requestor_address", "type": "felt"},
            {"name": "seed", "type": "felt"},
            {"name": "minimum_block_number", "type": "felt"},
            {"name": "random_words_len", "type": "felt"},
            {"name": "random_words", "type": "felt*"},
            {"name": "proof_len", "type": "felt"},
            {"name": "proof", "type": "felt*"},
        ],
        "keys": [],
        "name": "Randomness__proof",
        "type": "event",
    },
    {
        "data": [
            {"name": "requestor_address", "type": "felt"},
            {"name": "request_id", "type": "felt"},
            {"name": "status", "type": "felt"},
        ],
        "keys": [],
        "name": "Randomness__status_change",
        "type": "event",
    },
    {
        "inputs": [
            {"name": "proxy_admin", "type": "felt"},
            {"name": "public_key", "type": "Uint256"},
        ],
        "name": "initializer",
        "outputs": [],
        "type": "function",
    },
    {
        "inputs": [
            {"name": "requestor_address", "type": "felt"},
            {"name": "request_id", "type": "felt"},
            {"name": "status", "type": "felt"},
        ],
        "name": "update_status",
        "outputs": [],
        "type": "function",
    },
    {
        "inputs": [
            {"name": "seed", "type": "felt"},
            {"name": "callback_address", "type": "felt"},
            {"name": "callback_gas_limit", "type": "felt"},
            {"name": "publish_delay", "type": "felt"},
            {"name": "num_words", "type": "felt"},
        ],
        "name": "request_random",
        "outputs": [{"name": "request_id", "type": "felt"}],
        "type": "function",
    },
    {
        "inputs": [
            {"name": "request_id", "type": "felt"},
            {"name": "requestor_address", "type": "felt"},
            {"name": "seed", "type": "felt"},
            {"name": "minimum_block_number", "type": "felt"},
            {"name": "callback_address", "type": "felt"},
            {"name": "callback_gas_limit", "type": "felt"},
            {"name": "num_words", "type": "felt"},
        ],
        "name": "cancel_random_request",
        "outputs": [],
        "type": "function",
    },
    {
        "inputs": [
            {"name": "request_id", "type": "felt"},
            {"name": "requestor_address", "type": "felt"},
            {"name": "seed", "type": "felt"},
            {"name": "minimum_block_number", "type": "felt"},
            {"name": "callback_address", "type": "felt"},
            {"name": "callback_gas_limit", "type": "felt"},
            {"name": "random_words_len", "type": "felt"},
            {"name": "random_words", "type": "felt*"},
            {"name": "block_hash", "type": "felt"},
            {"name": "proof_len", "type": "felt"},
            {"name": "proof", "type": "felt*"},
        ],
        "name": "submit_random",
        "outputs": [],
        "type": "function",
    },
    {
        "inputs": [{"name": "new_implementation", "type": "felt"}],
        "name": "upgrade",
        "outputs": [],
        "type": "function",
    },
    {
        "inputs": [],
        "name": "get_implementation_hash",
        "outputs": [{"name": "address", "type": "felt"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [
            {"name": "requestor_address", "type": "felt"},
            {"name": "offset", "type": "felt"},
            {"name": "max_len", "type": "felt"},
            {"name": "request_ids_len", "type": "felt"},
            {"name": "request_ids", "type": "felt*"},
        ],
        "name": "get_pending_requests",
        "outputs": [
            {"name": "requests_len", "type": "felt"},
            {"name": "requests", "type": "felt*"},
        ],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [
            {"name": "requestor_address", "type": "felt"},
            {"name": "request_id", "type": "felt"},
        ],
        "name": "get_request_status",
        "outputs": [{"name": "status_", "type": "felt"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [{"name": "requestor_address", "type": "felt"}],
        "name": "requestor_current_index",
        "outputs": [{"name": "idx", "type": "felt"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [{"name": "requestor_address", "type": "felt"}],
        "name": "get_public_key",
        "outputs": [{"name": "pk", "type": "Uint256"}],
        "stateMutability": "view",
        "type": "function",
    },
]



import asyncio
import logging
from typing import List, Union

import requests
from aiohttp import ClientSession
from empiric.core.entry import SpotEntry
from empiric.core.utils import currency_pair_to_pair_id
from empiric.publisher.assets import EmpiricAsset, EmpiricSpotAsset
from empiric.publisher.types import PublisherFetchError, PublisherInterfaceT

logger = logging.getLogger(__name__)


class KaikoFetcher(PublisherInterfaceT):
    BASE_URL: str = "https://us.market-api.kaiko.io/v2/data/trades.v1/spot_direct_exchange_rate"
    SOURCE: str = "KAIKO"
    payload = {
        'interval': '2m',
        'page_size': '1',
    }

    publisher: str

    def __init__(self, assets: List[EmpiricAsset], publisher, api_key: str):
        self.assets = assets
        self.publisher = publisher
        self.headers = {"X-Api-Key": api_key}

    async def _fetch_pair(
        self, asset: EmpiricSpotAsset, session: ClientSession
    ) -> Union[SpotEntry, PublisherFetchError]:
        pair = asset["pair"]
        url = f"{self.BASE_URL}/{pair[0].lower()}/{pair[1].lower()}"
        async with session.get(url, headers=self.headers, params=self.payload) as resp:
            if resp.status == 404:
                return PublisherFetchError(
                    f"No data found for {'/'.join(pair)} from Kaiko"
                )
            result = await resp.json(content_type="application/json")
            if "error" in result and result["error"] == "Invalid Symbols Pair":
                return PublisherFetchError(
                    f"No data found for {'/'.join(pair)} from Kaiko"
                )

            return self._construct(asset, result)

    def _fetch_pair_sync(
        self, asset: EmpiricSpotAsset
    ) -> Union[SpotEntry, PublisherFetchError]:
        pair = asset["pair"]
        url = f"{self.BASE_URL}/{pair[0].lower()}/{pair[1].lower()}"

        resp = requests.get(url, headers=self.headers, params=self.payload)
        if resp.status_code == 404:
            return PublisherFetchError(f"No data found for {'/'.join(pair)} from Kaiko")
        result = resp.json(content_type="application/json")
        if "error" in result and result["error"] == "Invalid Symbols Pair":
            return PublisherFetchError(f"No data found for {'/'.join(pair)} from Kaiko")

        return self._construct(asset, result)

    async def fetch(
        self, session: ClientSession
    ) -> List[Union[SpotEntry, PublisherFetchError]]:
        entries = []
        for asset in self.assets:
            if asset["type"] != "SPOT":
                logger.debug(f"Skipping Kaiko for non-spot asset {asset}")
                continue
            entries.append(asyncio.ensure_future(self._fetch_pair(asset, session)))
        return await asyncio.gather(*entries, return_exceptions=True)

    def fetch_sync(self) -> List[Union[SpotEntry, PublisherFetchError]]:
        entries = []
        for asset in self.assets:
            if asset["type"] != "SPOT":
                logger.debug(f"Skipping Kaiko for non-spot asset {asset}")
                continue
            entries.append(self._fetch_pair_sync(asset))
        return entries

    def _construct(self, asset, result) -> SpotEntry:
        data = result["data"][0]
        pair = asset["pair"]

        timestamp = int(int(data["timestamp"]) / 1000)
        price = float(data["price"])
        price_int = int(price * (10 ** asset["decimals"]))
        pair_id = currency_pair_to_pair_id(*pair)

        logger.info(f"Fetched price {price} for {'/'.join(pair)} from Kaiko")

        return SpotEntry(
            pair_id=pair_id,
            price=price_int,
            timestamp=timestamp,
            source=self.SOURCE,
            publisher=self.publisher,
        )



PROXY_ABI = [
    {
        "data": [{"name": "implementation", "type": "felt"}],
        "keys": [],
        "name": "Upgraded",
        "type": "event",
    },
    {
        "data": [
            {"name": "previousAdminAddress", "type": "felt"},
            {"name": "newAdminAddress", "type": "felt"},
        ],
        "keys": [],
        "name": "AdminAddressChanged",
        "type": "event",
    },
    {
        "inputs": [
            {"name": "implementation_hash", "type": "felt"},
            {"name": "selector", "type": "felt"},
            {"name": "calldata_len", "type": "felt"},
            {"name": "calldata", "type": "felt*"},
        ],
        "name": "constructor",
        "outputs": [],
        "type": "constructor",
    },
    {
        "inputs": [
            {"name": "selector", "type": "felt"},
            {"name": "calldata_size", "type": "felt"},
            {"name": "calldata", "type": "felt*"},
        ],
        "name": "__default__",
        "outputs": [
            {"name": "retdata_size", "type": "felt"},
            {"name": "retdata", "type": "felt*"},
        ],
        "type": "function",
    },
    {
        "inputs": [
            {"name": "selector", "type": "felt"},
            {"name": "calldata_size", "type": "felt"},
            {"name": "calldata", "type": "felt*"},
        ],
        "name": "__l1_default__",
        "outputs": [],
        "type": "l1_handler",
    },
]



import asyncio
import logging
from typing import List, Union

import requests
from aiohttp import ClientSession
from empiric.core.entry import SpotEntry
from empiric.core.utils import currency_pair_to_pair_id
from empiric.publisher.assets import EmpiricAsset, EmpiricSpotAsset
from empiric.publisher.types import PublisherFetchError, PublisherInterfaceT

logger = logging.getLogger(__name__)


class OkxFetcher(PublisherInterfaceT):
    BASE_URL: str = "https://okx.com/api/v5/market/ticker"
    SOURCE: str = "OKX"

    publisher: str

    def __init__(self, assets: List[EmpiricAsset], publisher):
        self.assets = assets
        self.publisher = publisher

    async def _fetch_pair(
        self, asset: EmpiricSpotAsset, session: ClientSession
    ) -> Union[SpotEntry, PublisherFetchError]:
        pair = asset["pair"]
        url = f"{self.BASE_URL}?instId={pair[0]}-{pair[1]}-SWAP"

        async with session.get(url) as resp:
            if resp.status == 404:
                return PublisherFetchError(
                    f"No data found for {'/'.join(pair)} from OKX"
                )
            result = await resp.json(content_type="application/json")
            if (
                result["code"] == "51001"
                or result["msg"] == "Instrument ID does not exist"
            ):
                return PublisherFetchError(
                    f"No data found for {'/'.join(pair)} from OKX"
                )

            return self._construct(asset, result)

    def _fetch_pair_sync(
        self, asset: EmpiricSpotAsset
    ) -> Union[SpotEntry, PublisherFetchError]:
        pair = asset["pair"]
        url = f"{self.BASE_URL}?instId={pair[0]}-{pair[1]}-SWAP"

        resp = requests.get(url)
        if resp.status_code == 404:
            return PublisherFetchError(f"No data found for {'/'.join(pair)} from OKX")
        result = resp.json(content_type="application/json")
        if result["code"] == "51001" or result["msg"] == "Instrument ID does not exist":
            return PublisherFetchError(f"No data found for {'/'.join(pair)} from OKX")

        return self._construct(asset, result)

    async def fetch(
        self, session: ClientSession
    ) -> List[Union[SpotEntry, PublisherFetchError]]:
        entries = []
        for asset in self.assets:
            if asset["type"] != "SPOT":
                logger.debug(f"Skipping OKX for non-spot asset {asset}")
                continue
            entries.append(asyncio.ensure_future(self._fetch_pair(asset, session)))
        return await asyncio.gather(*entries, return_exceptions=True)

    def fetch_sync(self) -> List[Union[SpotEntry, PublisherFetchError]]:
        entries = []
        for asset in self.assets:
            if asset["type"] != "SPOT":
                logger.debug(f"Skipping OKX for non-spot asset {asset}")
                continue
            entries.append(self._fetch_pair_sync(asset))
        return entries

    def _construct(self, asset, result) -> SpotEntry:
        pair = asset["pair"]
        data = result["data"][0]

        timestamp = int(data["ts"])
        price = float(data["last"])
        price_int = int(price * (10 ** asset["decimals"]))
        pair_id = currency_pair_to_pair_id(*pair)
        volume = float(data["volCcy24h"])
        volume_int = int(volume)

        logger.info(f"Fetched price {price} for {'/'.join(pair)} from OKX")

        return SpotEntry(
            pair_id=pair_id,
            price=price_int,
            volume=volume_int,
            timestamp=timestamp,
            source=self.SOURCE,
            publisher=self.publisher,
        )



import asyncio
import logging
from typing import List, Union

import requests
from aiohttp import ClientSession
from empiric.core.entry import FutureEntry
from empiric.core.utils import currency_pair_to_pair_id
from empiric.publisher.assets import EmpiricAsset, EmpiricFutureAsset
from empiric.publisher.types import PublisherFetchError, PublisherInterfaceT
from datetime import datetime, timezone, timedelta

logger = logging.getLogger(__name__)


class BinanceFutureFetcher(PublisherInterfaceT):
    BASE_URL: str = "https://fapi.binance.com/fapi/v1/premiumIndex"
    VOLUME_URL: str = "https://fapi.binance.com/fapi/v1/ticker/24hr"
    SOURCE: str = "BINANCE"

    publisher: str

    def __init__(self, assets: List[EmpiricAsset], publisher):
        self.assets = assets
        self.publisher = publisher

    async def fetch_volume(self, asset, session):
        pair = asset["pair"]
        url = f"{self.VOLUME_URL}"
        selection = f"{pair[0]}{pair[1]}"
        volume_arr = []
        async with session.get(url) as resp:
            if resp.status == 404:
                return PublisherFetchError(
                    f"No data found for {'/'.join(pair)} from Binance"
                )
            result = await resp.json(content_type="application/json")
            for element in result: 
                if selection in element['symbol']:
                    volume_arr.append((element['symbol'],element['volume']))
            return volume_arr
    
    def fetch_volume_sync(self, asset):
        pair = asset["pair"]
        url = f"{self.VOLUME_URL}"
        selection = f"{pair[0]}{pair[1]}"
        volume_arr = []
        resp = requests.get(url)
        if resp.status_code == 404:
            return PublisherFetchError(
                f"No data found for {'/'.join(pair)} from Binance"
            )
        result = resp.json()
        for element in result: 
            if selection in element['symbol']:
                volume_arr.append((element['symbol'],element['volume']))
        return volume_arr
    
    async def _fetch_pair(
        self, asset: EmpiricFutureAsset, session: ClientSession
    ) -> Union[FutureEntry, PublisherFetchError]:
        pair = asset["pair"]
        filtered_data = []
        url = f"{self.BASE_URL}"
        selection = f"{pair[0]}{pair[1]}"
        async with session.get(url) as resp:
            if resp.status == 404:
                return PublisherFetchError(
                    f"No data found for {'/'.join(pair)} from Binance"
                )
            result = await resp.json(content_type="application/json")
            for element in result:
                if selection in element['symbol']:
                    filtered_data.append(element)
            volume_arr = await self.fetch_volume(asset, session)
            return self._construct(asset, filtered_data, volume_arr)

    def _fetch_pair_sync(
        self, asset: EmpiricFutureAsset
    ) -> Union[FutureEntry, PublisherFetchError]:
        pair = asset["pair"]
        url = f"{self.BASE_URL}"
        selection = f"{pair[0]}{pair[1]}"
        resp = requests.get(url)
        filtered_data = []
        if resp.status_code == 404:
            return PublisherFetchError(f"No data found for {'/'.join(pair)} from Binance")
        result = resp.json(content_type="application/json")
        if result["code"] == "51001" or result["msg"] == "Instrument ID does not exist":
            return PublisherFetchError(f"No data found for {'/'.join(pair)} from Binance")
        data_arr = result['symbols']
        for element in data_arr:
            if selection in element['symbol']:
                filtered_data.append(element)
        volume_arr = self.fetch_volume_sync(asset)
        return self._construct(asset, filtered_data, volume_arr)

    def fetch_sync(self):
        entries = []
        for asset in self.assets:
            if asset["type"] != "FUTURE":
                logger.debug(f"Skipping Binance for non-future asset {asset}")
                continue
            future_entries = self._fetch_pair_sync(asset)
            if isinstance(future_entries, list):
                entries.extend(future_entries)
            else:
                entries.append(future_entries)
        return entries

    async def fetch(self, session: ClientSession):
        entries = []
        for asset in self.assets:
            if asset["type"] != "FUTURE":
                logger.debug(f"Skipping Binance for non-future asset {asset}")
                continue
            future_entries = await self._fetch_pair(asset, session)
            if isinstance(future_entries, list):
                entries.extend(future_entries)
            else:
                entries.append(future_entries)
        return entries

    def retreive_volume(self,asset, volume_arr): 
        for list_asset, list_vol in volume_arr: 
            if asset == list_asset: 
                return list_vol
        return 0
    
    def _construct(self, asset, result, volume_arr) -> List[FutureEntry]:
        pair = asset["pair"]
        result_len = len(result)
        selection = f"{pair[0]}{pair[1]}"
        result_arr = []
        for i in range(0, result_len): 
            data = result[i]
            timestamp = int(data["time"])
            price = float(data["markPrice"])
            price_int = int(price * (10 ** asset["decimals"]))
            pair_id = currency_pair_to_pair_id(*pair)
            volume = float(self.retreive_volume(data['symbol'], volume_arr))
            volume_int = int(volume)
            if data['symbol']==selection: 
                expiry_timestamp = 0
            else:
                date_part = data['symbol'].split('_')[1]
                expiry_date = datetime.strptime(date_part, '%y%m%d')
                expiry_date = expiry_date.replace(hour=8, minute=0, second=0, tzinfo=timezone.utc)
                expiry_timestamp = int(expiry_date.timestamp())
            result_arr.append(FutureEntry(
                pair_id=pair_id,
                price=price_int,
                volume=volume_int,
                timestamp=int(timestamp/1000),
                source=self.SOURCE,
                publisher=self.publisher,
                expiry_timestamp=expiry_timestamp*1000,
            ))
        return result_arr



from __future__ import annotations

import abc
from typing import Dict, List, Optional, Tuple, Union

from empiric.core.utils import felt_to_str, str_to_felt
from web3 import Web3

class Entry(abc.ABC):
    @abc.abstractmethod
    def serialize(self) -> Dict[str, str]:
        ...

    @abc.abstractmethod
    def to_tuple(self) -> Dict[str, str]:
        ...

    def serialize_entries(self, entries: List[Entry]):
        serialized_entries = [
            entry.serialize() for entry in entries if issubclass(entry, Entry)
        ]
        return list(filter(lambda item: item is not None, serialized_entries))

    @staticmethod
    def flatten_entries(entries: List[SpotEntry]) -> List[int]:
        """This flattens entriees to tuples.  Useful when you need the raw felt array"""
        expanded = [entry.to_tuple() for entry in entries]
        flattened = [x for entry in expanded for x in entry]
        return [len(entries)] + flattened


class BaseEntry:
    timestamp: int
    source: int
    publisher: int

    def __init__(
        self, timestamp: int, source: Union[str, int], publisher: Union[str, int]
    ):
        if type(publisher) == str:
            publisher = str_to_felt(publisher)

        if type(source) == str:
            source = str_to_felt(source)

        self.timestamp = timestamp
        self.source = source
        self.publisher = publisher


class GenericEntry(Entry):
    base: BaseEntry
    key: int
    value: int

    def __init__(
        self,
        timestamp: int,
        source: Union[str, int],
        publisher: Union[str, int],
        key: Union[str, int],
        value: int,
    ):
        if type(publisher) == str:
            publisher = str_to_felt(publisher)

        if type(source) == str:
            source = str_to_felt(source)

        if type(key) == str:
            key = str_to_felt(key)

        self.base = BaseEntry(timestamp, source, publisher)
        self.key = key
        self.value = value

    def serialize(self) -> Dict[str, str]:
        return {
            "base": {
                "timestamp": self.base.timestamp,
                "source": self.base.source,
                "publisher": self.base.publisher,
            },
            "key": self.key,
            "value": self.value,
        }

    def to_tuple(self):
        return (
            self.base.timestamp,
            self.base.source,
            self.base.publisher,
            self.key,
            self.value,
        )

    def __repr__(self):
        return f'GenericEntry(key="{felt_to_str(self.key)}", value={self.value}, timestamp={self.base.timestamp}, source="{felt_to_str(self.base.source)}", publisher="{felt_to_str(self.base.publisher)}")'


class SpotEntry(Entry):
    base: BaseEntry
    pair_id: int
    price: int
    volume: int

    def __init__(
        self,
        pair_id: Union[str, int],
        price: int,
        timestamp: int,
        source: Union[str, int],
        publisher: Union[str, int],
        volume: Optional[int] = 0,
    ) -> None:
        # TODO: This should be network agnostic
        if type(pair_id) == str:
            pair_id = str_to_felt(pair_id)

        if type(publisher) == str:
            publisher = str_to_felt(publisher)

        if type(source) == str:
            source = str_to_felt(source)

        self.base = BaseEntry(timestamp, source, publisher)
        self.pair_id = pair_id
        self.price = price
        self.volume = volume

    def __eq__(self, other):
        if isinstance(other, SpotEntry):
            return (
                self.pair_id == other.pair_id
                and self.price == other.price
                and self.base.timestamp == other.base.timestamp
                and self.base.source == other.base.source
                and self.base.publisher == other.base.publisher
                and self.volume == other.volume
            )
        # This supports comparing against entries that are returned by starknet.py,
        # which will be namedtuples.
        if isinstance(other, Tuple) and len(other) == 4:
            return (
                self.pair_id == other.pair_id
                and self.price == other.price
                and self.base.timestamp == other.base.timestamp
                and self.base.source == other.base.source
                and self.base.publisher == other.base.publisher
                and self.volume == other.volume
            )
        return False

    def to_tuple(self):
        return (
            self.base.timestamp,
            self.base.source,
            self.base.publisher,
            self.pair_id,
            self.price,
            self.volume,
        )

    def serialize(self) -> Dict[str, str]:
        return {
            "base": {
                "timestamp": self.base.timestamp,
                "source": self.base.source,
                "publisher": self.base.publisher,
            },
            "pair_id": self.pair_id,
            "price": self.price,
            "volume": self.volume,
        }

    def serialize_evm(self) -> Dict[str, str]:
        return {
            "base": {
                "timestamp": self.base.timestamp,
                "source": Web3.toBytes(text=felt_to_str(self.base.source)),
                "publisher": Web3.toBytes(text=felt_to_str(self.base.publisher)),
            },
            "pairId": Web3.toBytes(text=felt_to_str(self.pair_id)),
            "price": self.price,
            "volume": self.volume,
        }

    @staticmethod
    def from_dict(entry_dict: Dict[str, str]) -> "SpotEntry":
        return SpotEntry(
            entry_dict["base"]["pair_id"],
            entry_dict["price"],
            entry_dict["timestamp"],
            entry_dict["base"]["source"],
            entry_dict["base"]["publisher"],
        )

    @staticmethod
    def serialize_entries(entries: List[SpotEntry]) -> List[Dict[str, int]]:
        """serialize entries to a List of dictionaries"""
        # TODO (rlkelly): log errors
        serialized_entries = [
            entry.serialize()
            for entry in entries
            # TODO (rlkelly): This needs to be much more resilient to publish errors
            if isinstance(entry, SpotEntry)
        ]
        return list(filter(lambda item: item is not None, serialized_entries))

    @staticmethod
    def serialize_entries_evm(entries: List[SpotEntry]) -> List[Dict[str, int]]:
        """serialize entries to a List of dictionaries"""
        # TODO (rlkelly): log errors
        serialized_entries = [
            entry.serialize_evm()
            for entry in entries
            # TODO (rlkelly): This needs to be much more resilient to publish errors
            if isinstance(entry, SpotEntry)
        ]
        return list(filter(lambda item: item is not None, serialized_entries))

    def __repr__(self):
        return f'SpotEntry(pair_id="{felt_to_str(self.pair_id)}", price={self.price}, timestamp={self.base.timestamp}, source="{felt_to_str(self.base.source)}", publisher="{felt_to_str(self.base.publisher)}, volume={self.volume})")'


class FutureEntry(Entry):
    base : BaseEntry
    pair_id: int
    price: int
    expiry_timestamp: int
    volume :int 

    def __init__(self, timestamp, source, publisher, pair_id, price, expiry_timestamp, volume):
        if type(pair_id) == str:
            pair_id = str_to_felt(pair_id)

        if type(publisher) == str:
            publisher = str_to_felt(publisher)

        if type(source) == str:
            source = str_to_felt(source)

        self.base = BaseEntry(timestamp, source, publisher)
        self.pair_id = pair_id
        self.price = price
        self.expiry_timestamp = expiry_timestamp
        self.volume = volume

    def __eq__(self, other):
        if isinstance(other, FutureEntry):
            return (
                self.pair_id == other.pair_id
                and self.price == other.price
                and self.base.timestamp == other.base.timestamp
                and self.base.source == other.base.source
                and self.base.publisher == other.base.publisher
                and self.expiry_timestamp == other.expiry_timestamp
                and self.volume == other.volume
            )
        # This supports comparing against entries that are returned by starknet.py,
        # which will be namedtuples.
        if isinstance(other, Tuple) and len(other) == 4:
            return (
                self.pair_id == other.pair_id
                and self.price == other.price
                and self.base.timestamp == other.base.timestamp
                and self.base.source == other.base.source
                and self.base.publisher == other.base.publisher
                and self.expiry_timestamp == other.expiry_timestamp
                and self.volume == other.volume
            )
        return False

    def to_tuple(self):
        return (
            self.base.timestamp,
            self.base.source,
            self.base.publisher,
            self.pair_id,
            self.price,
            self.expiry_timestamp,
            self.volume,
        )

    def serialize(self) -> Dict[str, str]:
        return {
            "base": {
                "timestamp": self.base.timestamp,
                "source": self.base.source,
                "publisher": self.base.publisher,
            },
            "pair_id": self.pair_id,
            "price": self.price,
            "expiry_timestamp": self.expiry_timestamp,
            "volume": self.volume,
        }

    @staticmethod
    def serialize_entries(entries: List[FutureEntry]) -> List[Dict[str, int]]:
        """serialize entries to a List of dictionaries"""
        # TODO (rlkelly): log errors
        serialized_entries = [
            entry.serialize()
            for entry in entries
            # TODO (rlkelly): This needs to be much more resilient to publish errors
            if isinstance(entry, FutureEntry)
        ]
        return list(filter(lambda item: item is not None, serialized_entries))



import collections
import logging
from typing import List, Optional

from empiric.core.contract import Contract
from empiric.core.entry import FutureEntry, SpotEntry
from empiric.core.types import AggregationMode
from empiric.core.utils import str_to_felt
from starknet_py.contract import InvokeResult
from starknet_py.net.account.account import Account
from starknet_py.net.client import Client

logger = logging.getLogger(__name__)

OracleResponse = collections.namedtuple(
    "OracleResponse",
    ["price", "decimals", "last_updated_timestamp", "num_sources_aggregated"],
)


class OracleMixin:
    publisher_registry: Contract
    client: Client
    account: Account

    async def publish_spot_entry(
        self,
        pair_id: int,
        value: int,
        timestamp: int,
        source: int,
        publisher: int,
        volume: int = 0,
        max_fee: int = int(1e18),
    ) -> InvokeResult:
        if not self.is_user_client:
            raise AttributeError(
                "Must set account.  You may do this by invoking self._setup_account_client(private_key, account_contract_address)"
            )
        invocation = await self.oracle.publish_spot_entry.invoke(
            {
                "pair_id": pair_id,
                "value": value,
                "timestamp": timestamp,
                "source": source,
                "publisher": publisher,
            },
            max_fee=max_fee,
        )
        return invocation

    async def publish_many(
        self,
        entries: List[any],
        pagination: Optional[int] = 40,
        max_fee=int(1e18),
    ) -> List[InvokeResult]:
        if len(entries) == 0:
            logger.warning("Skipping publishing as entries array is empty")
            return

        invocations = []
        serialized_spot_entries = SpotEntry.serialize_entries(entries)
        if pagination:
            ix = 0
            while ix < len(serialized_spot_entries):
                entries_subset = serialized_spot_entries[ix : ix + pagination]
                invocation = await self.oracle.publish_spot_entries.invoke(
                    entries_subset,
                    callback=self.track_nonce,
                    max_fee=max_fee,
                )
                ix += pagination
                invocations.append(invocation)
                logger.debug(str(invocation))
                logger.info(
                    f"Sent {len(entries_subset)} updated spot entries with transaction {hex(invocation.hash)}"
                )
        elif len(serialized_spot_entries) > 0:
            invocation = await self.oracle.publish_spot_entries.invoke(
                serialized_spot_entries, max_fee=max_fee
            )
            invocations.append(invocation)
            logger.debug(str(invocation))
            logger.info(
                f"Sent {len(serialized_spot_entries)} updated spot entries with transaction {hex(invocation.hash)}"
            )

        serialized_future_entries = FutureEntry.serialize_entries(entries)
        if pagination:
            ix = 0
            while ix < len(serialized_future_entries):
                entries_subset = serialized_future_entries[ix : ix + pagination]
                invocation = await self.oracle.publish_future_entries.invoke(
                    entries_subset,
                    callback=self.track_nonce,
                    max_fee=max_fee,
                )
                ix += pagination
                invocations.append(invocation)
                logger.debug(str(invocation))
                logger.info(
                    f"Sent {len(entries_subset)} updated future entries with transaction {hex(invocation.hash)}"
                )
        elif len(serialized_future_entries) > 0:
            invocation = await self.oracle.publish_future_entries.invoke(
                serialized_future_entries, max_fee=max_fee
            )
            invocations.append(invocation)
            logger.debug(str(invocation))
            logger.info(
                f"Sent {len(serialized_future_entries)} updated future entries with transaction {hex(invocation.hash)}"
            )

        return invocations

    async def get_entries(self, pair_id, sources=[]) -> List[SpotEntry]:
        if isinstance(pair_id, str):
            pair_id = str_to_felt(pair_id)
        elif not isinstance(pair_id, int):
            raise TypeError(
                "Pair ID must be string (will be converted to felt) or integer"
            )
        response = await self.oracle.get_spot_entries.call(pair_id, sources)

        return [SpotEntry.from_dict(entry) for entry in response.entries]

    async def get_spot(
        self,
        pair_id,
        aggregation_mode: AggregationMode = AggregationMode.MEDIAN,
        sources=None,
    ) -> OracleResponse:
        if isinstance(pair_id, str):
            pair_id = str_to_felt(pair_id)
        elif not isinstance(pair_id, int):
            raise TypeError(
                "Pair ID must be string (will be converted to felt) or integer"
            )
        if sources is None:
            response = await self.oracle.get_spot.call(
                pair_id,
                aggregation_mode.value,
            )
        else:
            response = await self.oracle.get_spot_for_sources.call(
                pair_id, aggregation_mode.value, sources
            )

        return OracleResponse(
            response.price,
            response.decimals,
            response.last_updated_timestamp,
            response.num_sources_aggregated,
        )

    async def get_future(
        self,
        pair_id,
        expiry_timestamp,
        aggregation_mode: AggregationMode = AggregationMode.MEDIAN,
        # TODO Add sources on the oracle contract and then in the client here
        # sources=None,
    ) -> OracleResponse:
        if isinstance(pair_id, str):
            pair_id = str_to_felt(pair_id)
        elif not isinstance(pair_id, int):
            raise TypeError(
                "Pair ID must be string (will be converted to felt) or integer"
            )

        response = await self.oracle.get_future.call(
            pair_id,
            expiry_timestamp,
            aggregation_mode.value,
        )

        return OracleResponse(
            response.price,
            response.decimals,
            response.last_updated_timestamp,
            response.num_sources_aggregated,
        )

    async def get_spot_decimals(self, pair_id) -> int:
        if isinstance(pair_id, str):
            pair_id = str_to_felt(pair_id)
        elif not isinstance(pair_id, int):
            raise TypeError(
                "Pair ID must be string (will be converted to felt) or integer"
            )

        response = await self.oracle.get_spot_decimals.call(
            pair_id,
        )

        return response

    async def set_checkpoint(
        self,
        pair_id: int,
        aggregation_mode: int = str_to_felt("MEDIAN"),
        max_fee=int(1e16),
    ) -> InvokeResult:
        if not self.is_user_client:
            raise AttributeError(
                "Must set account.  You may do this by invoking self._setup_account_client(private_key, account_contract_address)"
            )
        invocation = await self.oracle.set_checkpoint.invoke(
            pair_id,
            aggregation_mode,
            callback=self.track_nonce,
            max_fee=max_fee,
        )
        return invocation

    async def set_future_checkpoint(
        self,
        pair_id: int,
        expiry_timestamp: int,
        aggregation_mode: int = str_to_felt("MEDIAN"),
        max_fee=int(1e16),
    ) -> InvokeResult:
        if not self.is_user_client:
            raise AttributeError(
                "Must set account.  You may do this by invoking self._setup_account_client(private_key, account_contract_address)"
            )
        invocation = await self.oracle.set_future_checkpoint.invoke(
            pair_id,
            expiry_timestamp,
            aggregation_mode,
            max_fee=max_fee,
        )
        return invocation

    async def set_future_checkpoints(
        self,
        pair_ids: List[int],
        expiry_timestamps: List[int],
        aggregation_mode: int = str_to_felt("MEDIAN"),
        max_fee=int(1e16),
        pagination: Optional[int] = 15,
    ) -> InvokeResult:
        if not self.is_user_client:
            raise AttributeError(
                "Must set account.  You may do this by invoking self._setup_account_client(private_key, account_contract_address)"
            )

        if pagination:
            ix = 0
            while ix < len(pair_ids):
                pair_ids_subset = pair_ids[ix : ix + pagination]
                invocation = await self.oracle.set_future_checkpoints.invoke(
                    pair_ids_subset,
                    expiry_timestamps,
                    aggregation_mode,
                    callback=self.track_nonce,
                    max_fee=max_fee,
                )
                ix += pagination
                logger.debug(str(invocation))
                logger.info(
                    f"Set future checkpoints for {len(pair_ids_subset)} pair IDs with transaction {hex(invocation.hash)}"
                )
        else:
            invocation = await self.oracle.set_future_checkpoints.invoke(
                pair_ids,
                expiry_timestamps,
                aggregation_mode,
                max_fee=max_fee,
            )

        return invocation

    async def set_checkpoints(
        self,
        pair_ids: List[int],
        aggregation_mode: int = str_to_felt("MEDIAN"),
        max_fee=int(1e18),
        pagination: Optional[int] = 15,
    ) -> InvokeResult:
        if not self.is_user_client:
            raise AttributeError(
                "Must set account.  You may do this by invoking self._setup_account_client(private_key, account_contract_address)"
            )
        if pagination:
            ix = 0
            while ix < len(pair_ids):
                pair_ids_subset = pair_ids[ix : ix + pagination]
                invocation = await self.oracle.set_checkpoints.invoke(
                    pair_ids_subset,
                    aggregation_mode,
                    # callback=self.track_nonce,
                    max_fee=max_fee,
                )
                ix += pagination
                logger.debug(str(invocation))
                logger.info(
                    f"Set checkpoints for {len(pair_ids_subset)} pair IDs with transaction {hex(invocation.hash)}"
                )
        else:
            invocation = await self.oracle.set_checkpoints.invoke(
                pair_ids,
                aggregation_mode,
                # callback=self.track_nonce,
                max_fee=max_fee,
            )

        return invocation
